// Generated by CoffeeScript 1.12.6
var DictionaryKnot, ListKnot, NumKnot, RefKnot, Semaphore, _, _ASAP, _proto, debug, duct, exports,
  hasProp = {}.hasOwnProperty,
  extend = function(child, parent) { for (var key in parent) { if (hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; },
  slice = [].slice;

_ = require('lodash');

duct = require('./duct');

Semaphore = require('./semaphore');

debug = require('debug')('knot');


/* 
Knot는 프로그램 처리 흐름상의 결절이다.
시작점이자 끝점으로 볼수있으며,
상태를 확인할수 있다.
내부 자료 타입에 따라서 3~4가지를 둔다.

  RefKnot - 무엇인가를 참조함 
  ListKnot - 내부가 Array, .push가 가능
  DictionaryKnot - 내부가 Dictionary(=Object), .set 가능
  NumberKnot - 내부가 Number, .inc(.increase) 가능
  
Knot 공통. = RefKnot
  결절 내부 공간에 대한 Get/Set
  .getStatus
  .setStatus

  특정 상황이 될때까지 대기 가능
  .wait = .waitOnce
    until: (knot)-> return false
    then : (knot)-> 

  wait는 한번만 처리하고 when은 반복 수행
  .when 
    if: (knot)-> return false
    then : (knot)-> 

  .manual : pullout을 그대로 호출함.


  reactivity 계열 - pullout호출을 자동화 시킴
    * noReact : 반응하지 마라.
    * consecution: 들어오는 즉시 나감. 처리함수의 연속.
    * asap : 가능한 빨리 큐에서 꺼냄. 비동기
    * debounce: 지연된 시간내의 것을 모아서.
    * throttle: 우선 들어온것 처리하고, given time 동안 처리를 지연하여 처리
    * interval: 지정된 간격으로 나감
    * backPressure:
        pullOut이 처리되면 연속 호출
        처리할 데이터가 없으면? setTimeout
        최초의 시작은? 자동으로 함.
  
  Puller 계열
    * 100% 커스텀.
    
  Outer 계열
    * serial 순차 처리
    * parallel 전체 동시 병렬 처리
    * nParallel 갯수 제한 동시 처리

  Handler
    * 100% 커스텀.

   
ListKnot() 
  .push 
  
  Puller 계열
    * all : 들어온 것을 그대로 내보냄
    * dequeue: 1개만 꺼냄
    * latest: 다 버리고 가장 마지막 것
    * reduce(init_acc, acc_fn): 리듀싱 함수 호출

DictionaryKnot 
  .set(path, value) 
  .get(path)
  Puller 계열
    * pairs : 모든 키,값의 쌍 [k,v]
    * pairObject: 모든 {key: 키값, value: 값} 구조
    * keys: 모든 Key만
    * values: 모든 value만
    * reduce(init_acc, acc_fn)  
  
  
    
NumKnot 
  .inc 
  
  Puller
    * current : 현재 값 읽기
    
 
대량 처리시의 문제.
이때는 puts로 전부 넣기가 무리고,
스트림처리식도 부족함. 동기화가 안되서, 막 밀어넣고 전부 버퍼링 되게됨.
BackPressure 개념뿐인데..
문제는 끝나는 시점이 정확하지 않다.
  끝 = .next() is null  + all pullout execute done

>> 해법
knot = RefKnot()
  .setStatus MongoCursor
  .backPressure()
   * .noPuller() 
   * .serial() / .parallel()
  
knot.puller.do (self)-> [ self.getStatus().next()  ]
knot.handler.do (data)-> ....
  .await (data, done)-> ...
 */

_ASAP = function(fn) {
  return setTimeout(fn, 0);
};

if ((typeof process !== "undefined" && process !== null ? process.nextTick : void 0) != null) {
  _ASAP = process.nextTick;
}

_proto = function(klass, dict) {
  var key, results, v;
  results = [];
  for (key in dict) {
    if (!hasProp.call(dict, key)) continue;
    v = dict[key];
    results.push(klass.prototype[key] = v);
  }
  return results;
};

RefKnot = (function() {
  function RefKnot() {
    this._triggers = [];
    this.status = null;
    this.reset();
    this.serial();
  }

  RefKnot.prototype.reset = function() {
    this.reactivity = duct();
    this.puller = duct();
    this.outer = duct();
    return this.handler = duct();
  };

  RefKnot.prototype.getStatus = function() {
    return this.status;
  };

  RefKnot.prototype.setStatus = function(status) {
    this.status = status;
    return this._trigging();
  };

  RefKnot.prototype._trigging = function() {
    var self;
    self = this;
    debug('check _trigging', this._triggers, this.constructor.name);
    return this._triggers = _.filter(this._triggers, function(cfg) {
      debug('call cfg.if', cfg["if"]);
      if (cfg["if"](self)) {
        _ASAP(function() {
          return cfg.then(self);
        });
        if (cfg.once === true) {
          return false;
        }
      }
      return true;
    });
  };

  RefKnot.prototype.wait = function(wait_cfg) {
    if (!wait_cfg["if"]) {
      wait_cfg["if"] = wait_cfg.until;
    }
    wait_cfg.once = true;
    return this._triggers.push(wait_cfg);
  };

  RefKnot.prototype.when = function(when_cfg) {
    return this._triggers.push(when_cfg);
  };

  RefKnot.prototype.pullOut = function(callback) {
    var _fn, self;
    self = this;
    _fn = duct().await("data", function(done) {
      debug('call puller');
      return self.puller(self, done);
    }).restore("data").await(function(data, done) {
      debug('call outter');
      return self.outer(data, done);
    })["do"](function(data) {
      debug('reactivity; after-pullout');
      return self.reactivity('after-pullout', data);
    });
    _fn(function(err) {
      if (callback) {
        return callback(err, self);
      }
    });
    return this;
  };

  RefKnot.prototype.setHandler = function(handler) {
    this.handler = handler;
    return this;
  };

  RefKnot.prototype.setDataHandler = function(handler) {
    this.handler = handler;
    return this;
  };

  RefKnot.prototype.buildHandler = function(fn) {
    fn(this.handler);
    return this;
  };

  RefKnot.prototype.doHandling = function(data, callback) {
    this.handler(data, callback);
    return this;
  };

  RefKnot.prototype.noReact = function() {
    var self;
    self = this;
    self.reactivity.clear();
    return self;
  };

  RefKnot.prototype.consecution = function() {
    var self;
    self = this;
    self.reactivity.clear()["do"](function(timing_name) {
      if (timing_name !== 'after-change') {
        return;
      }
      debug('consecution call pullOut');
      return self.pullOut();
    });
    return self;
  };

  RefKnot.prototype.asap = function() {
    var self;
    self = this;
    self.reactivity.clear()["do"](function(timing_name) {
      if (timing_name !== 'after-change') {
        return;
      }
      return _ASAP(function() {
        return self.pullOut();
      });
    });
    return self;
  };

  RefKnot.prototype.backPressure = function(msec) {
    var self;
    self = this;
    return self.reactivity.clear()["do"](function(timing_name, data) {
      var _pullout;
      if (timing_name !== 'after-pullout') {
        return;
      }
      _pullout = function() {
        return self.pullOut();
      };
      if (data.length > 0) {
        return _ASAP(_pullout);
      } else {
        return setTimeout(_pullout, msec);
      }
    });
  };

  RefKnot.prototype.interval = function(msec) {
    var _tick, self;
    self = this;
    self.reactivity.clear();
    _tick = function() {
      return self.pullOut();
    };
    setInterval(_tick, msec);
    return self;
  };

  RefKnot.prototype.throttle = function(msec) {
    var self;
    self = this;
    self.reactivity.clear()["do"](function(timing_name) {
      var _dfn;
      if (timing_name !== 'after-change') {
        return;
      }
      if (self.tid) {
        return;
      }
      _dfn = function() {
        self.tid = null;
        return self.pullOut();
      };
      self.tid = setTimeout(_dfn, msec);
      return self.pullOut();
    });
    return self;
  };

  RefKnot.prototype.debounce = function(msec) {
    var self;
    self = this;
    self.reactivity.clear()["do"](function(timing_name) {
      var _dfn;
      if (timing_name !== 'after-change') {
        return;
      }
      if (self.tid) {
        return;
      }
      _dfn = function() {
        self.tid = null;
        return self.pullOut();
      };
      return self.tid = setTimeout(_dfn, msec);
    });
    return self;
  };

  RefKnot.prototype.par = function() {
    return this.parallel();
  };

  RefKnot.prototype.parallel = function() {
    var self;
    self = this;
    self.outer.clear().await(function(data, done) {
      var _fn;
      _fn = duct();
      _.forEach(data, function(datum, inx) {
        return _fn.async(inx, function(done) {
          return self.doHandling(datum, done);
        });
      });
      _fn.wait();
      return _fn(done);
    });
    return self;
  };

  RefKnot.prototype.ser = function() {
    return this.serial();
  };

  RefKnot.prototype.serial = function() {
    var self;
    self = this;
    self.outer.clear().await(function(data, done) {
      var _fn;
      _fn = duct();
      _.forEach(data, function(datum, inx) {
        return _fn.await(inx, function(done) {
          return self.doHandling(datum, done);
        });
      });
      return _fn(done);
    });
    return self;
  };

  RefKnot.prototype.nPar = function(concurrent) {
    return this.nParallel(concurrent);
  };

  RefKnot.prototype.nParallel = function(concurrent) {
    var s, self;
    self = this;
    s = new Semaphore(concurrent);
    self.outer.clear().await(function(data, done) {
      var _fn;
      _fn = duct();
      _.forEach(data, function(datum, inx) {
        return _fn.async(inx, function(done) {
          return s.enter(function() {
            return self.doHandling(datum, function(err) {
              s.leave();
              return done(err);
            });
          });
        });
      });
      _fn.wait();
      return _fn(done);
    });
    return self;
  };

  return RefKnot;

})();

ListKnot = (function(superClass) {
  extend(ListKnot, superClass);

  function ListKnot() {
    ListKnot.__super__.constructor.call(this);
    this.setStatus([]);
    this.all();
  }

  ListKnot.prototype.push = function() {
    var args, ref;
    args = 1 <= arguments.length ? slice.call(arguments, 0) : [];
    (ref = this.getStatus()).push.apply(ref, args);
    this.reactivity("after-change");
    this._trigging();
    return this;
  };

  ListKnot.prototype.pushAll = function(list) {
    var ref;
    (ref = this.getStatus()).push.apply(ref, list);
    this.reactivity("after-change");
    this._trigging();
    return this;
  };

  ListKnot.prototype.all = function() {
    var self;
    self = this;
    self.puller.clear()["do"](function(knot) {
      var list;
      list = knot.getStatus();
      knot.setStatus([]);
      return this.feedback.reset(list);
    });
    return self;
  };

  ListKnot.prototype.latest = function() {
    var self;
    self = this;
    self.puller.clear()["do"](function(knot) {
      var list;
      list = knot.getStatus();
      knot.setStatus([]);
      return this.feedback.reset([_.last(list)]);
    });
    return self;
  };

  ListKnot.prototype.reduce = function(reduce_fn) {
    var self;
    self = this;
    self.puller.clear()["do"](function(knot) {
      var list, val;
      list = knot.getStatus();
      knot.setStatus([]);
      val = reduce_fn(list);
      return this.feedback.reset([val]);
    });
    return self;
  };

  return ListKnot;

})(RefKnot);

DictionaryKnot = (function(superClass) {
  extend(DictionaryKnot, superClass);

  function DictionaryKnot() {
    DictionaryKnot.__super__.constructor.call(this);
    this.setStatus({});
  }

  DictionaryKnot.prototype.set = function(key, value) {
    _.set(this.getStatus(), key, value);
    this.reactivity("after-change");
    return this._trigging();
  };

  DictionaryKnot.prototype.get = function(key, value) {
    return _.get(this.getStatus(), key);
  };

  return DictionaryKnot;

})(RefKnot);

NumKnot = (function(superClass) {
  extend(NumKnot, superClass);

  function NumKnot() {
    NumKnot.__super__.constructor.call(this);
    this.setStatus(0);
  }

  NumKnot.prototype.inc = function(diff) {
    if (diff == null) {
      diff = 1;
    }
    debug('inc', diff);
    return this.setStatus(this.getStatus() + diff);
  };

  NumKnot.prototype.current = function() {
    var self;
    self = this;
    self.puller.clear()["do"](function(self) {
      var val;
      val = self.getStatus();
      return this.feedback.reset([val]);
    });
    return self;
  };

  return NumKnot;

})(RefKnot);

module.exports = exports = {
  Ref: RefKnot,
  List: ListKnot,
  Num: NumKnot,
  Dictionary: DictionaryKnot
};
